program Example4;

uses
  Windows,
  Messages,
  math,
  dglOpenGL,
  DglShader,
  logger;

const
  WND_TITLE = 'OpenGL 3.x Demo';
  FPS_TIMER = 1;                     // Timer to calculate FPS
  FPS_INTERVAL = 1000;               // Calculate FPS every 1000 ms

var
  glminorversion: integer;
  glmajorversion: integer;
  glversion: string;

  winwidth: integer = 600;           // window x size
  winheight: integer = 400;          // window y size

  h_Wnd  : HWND;                     // Global window handle
  h_DC   : HDC;                      // Global device context
  h_RC   : HGLRC;                    // OpenGL rendering context
  h_RC3  : HGLRC;                    // OpenGL 3.x rendering context
  keys : Array[0..255] of Boolean;   // Holds keystrokes
  FPSCount : Integer = 0;            // Counter for FPS
  ElapsedTime : Integer;             // Elapsed time between frames

  // User vaiables
  xSpeed, ySpeed : glFloat;

  // Mouse Variables
  MouseMove : boolean = false;        // Is the mouse moving?
  MoveXcoord, MoveYcoord : GLfloat;   // Current X and Y position of the cursor

  MouseButton : Integer = -1;         // mouse button down
  Xcoord, Ycoord : Integer;           // Current X and Y position where the mouse was clicked

  OpenGLInitialized : Boolean;
  OpenGL3Initialized : Boolean;       // OpenGL 3.x initialized

  //shaders
  glslsimplevert: TGLSLShader;
  glslsimplefrag: TGLSLShader;
  glslsimpleprog: TGLSLProgram;

  g_projectionLocation: GLint;
  g_indices: GLuint; //The VBO for the indices.
  vao: Gluint;

  type
    GLMatrix= array[0..15] of GLfloat;
    GLvect= array[0..2] of GLfloat;

  const
  // Points of a cube.
  points: array[0..95] of GLFloat=
             (-0.5, -0.5, -0.5, 1.0,
						  -0.5, -0.5, +0.5, 1.0,
						  +0.5, -0.5, +0.5, 1.0,
						  +0.5, -0.5, -0.5, 1.0,
						  -0.5, +0.5, -0.5, 1.0,
						  -0.5, +0.5, +0.5, 1.0,
						  +0.5, +0.5, +0.5, 1.0,
						  +0.5, +0.5, -0.5, 1.0,
						  -0.5, -0.5, -0.5, 1.0,
						  -0.5, +0.5, -0.5, 1.0,
						  +0.5, +0.5, -0.5, 1.0,
						  +0.5, -0.5, -0.5, 1.0,
						  -0.5, -0.5, +0.5, 1.0,
						  -0.5, +0.5, +0.5, 1.0,
						  +0.5, +0.5, +0.5, 1.0,
						  +0.5, -0.5, +0.5, 1.0,
						  -0.5, -0.5, -0.5, 1.0,
						  -0.5, -0.5, +0.5, 1.0,
						  -0.5, +0.5, +0.5, 1.0,
						  -0.5, +0.5, -0.5, 1.0,
						  +0.5, -0.5, -0.5, 1.0,
						  +0.5, -0.5, +0.5, 1.0,
						  +0.5, +0.5, +0.5, 1.0,
						  +0.5, +0.5, -0.5, 1.0);
  // Normals of a cube.
  normals: array[0..71] of GLFloat=
             (+0.0, -1.0, +0.0,
						  +0.0, -1.0, +0.0,
						  +0.0, -1.0, +0.0,
						  +0.0, -1.0, +0.0,
						  +0.0, +1.0, +0.0,
						  +0.0, +1.0, +0.0,
						  +0.0, +1.0, +0.0,
						  +0.0, +1.0, +0.0,
						  +0.0, +0.0, -1.0,
						  +0.0, +0.0, -1.0,
						  +0.0, +0.0, -1.0,
						  +0.0, +0.0, -1.0,
						  +0.0, +0.0, +1.0,
						  +0.0, +0.0, +1.0,
						  +0.0, +0.0, +1.0,
						  +0.0, +0.0, +1.0,
						  -1.0, +0.0, +0.0,
						  -1.0, +0.0, +0.0,
						  -1.0, +0.0, +0.0,
						  -1.0, +0.0, +0.0,
						  +1.0, +0.0, +0.0,
						  +1.0, +0.0, +0.0,
						  +1.0, +0.0, +0.0,
						  +1.0, +0.0, +0.0);
  // The associated indices.
  indices: array[0..35] of GLInt=
               (0, 2, 1,
								0, 3, 2,
								4, 5, 6,
								4, 6, 7,
								8, 9, 10,
								8, 10, 11,
								12, 15, 14,
								12, 14, 13,
								16, 17, 18,
								16, 18, 19,
								20, 23, 22,
								20, 22, 21);

{$R *.RES}

{------------------------------------------------------------------}
{  Function to convert int to string. (No sysutils = smaller EXE)  }
{------------------------------------------------------------------}
function IntToStr(Num : Integer) : String;  // using SysUtils increase file size by 100K
begin
  Str(Num, result);
end;

function amount(vector: GLvect): GLfloat;
begin
  result := sqrt(vector[0] * vector[0] + vector[1] * vector[1] + vector[2] * vector[2]);
end;

procedure normalize(var vector: GLvect);
var
  a: GLfloat;
  i: integer;
begin
	a := amount(vector);

  if (a = 0.0) then
	begin
    	exit;
	end;

	for i := 0 to 2 do
	begin
    	vector[i] := vector[i] / a;
	end;
end;

function dotProduct(const vector0: GLvect; const vector1: GLvect): GLfloat;
begin
	result := vector0[0]*vector1[0] + vector0[1]*vector1[1] + vector0[2]*vector1[2];
end;

procedure crossProduct(var aresult: GLvect; const vector0: GLvect; const vector1: GLvect);
var
  temp: GLvect;
  i: integer;
begin
	temp[0] := vector0[1]*vector1[2] - vector0[2]*vector1[1];
	temp[1] := vector0[2]*vector1[0] - vector0[0]*vector1[2];
	temp[2] := vector0[0]*vector1[1] - vector0[1]*vector1[0];

	for i := 0 to 2 do
	begin
    	aresult[i] := temp[i];
	end;
end;

procedure loadIdentity(var matrix: GLmatrix);
begin
	matrix[0] := 1.0;
	matrix[1] := 0.0;
	matrix[2] := 0.0;
	matrix[3] := 0.0;

	matrix[4] := 0.0;
	matrix[5] := 1.0;
	matrix[6] := 0.0;
	matrix[7] := 0.0;

	matrix[8] := 0.0;
	matrix[9] := 0.0;
	matrix[10] := 1.0;
	matrix[11] := 0.0;

	matrix[12] := 0.0;
	matrix[13] := 0.0;
	matrix[14] := 0.0;
	matrix[15] := 1.0;
end;

procedure multMatrix(var aresult: GLmatrix; const matrix0: GLmatrix; const matrix1: GLmatrix);
var
  i, k: integer;
  temp: GLmatrix;
begin
	for i := 0 to 15 do
	begin
		temp[i] := 0.0;

		for k := 0 to 3 do
		begin
			//			  		row   column   		   row column
			temp[i] := temp[i]+ matrix0[(i mod 4)+(k*4)] * matrix1[k+((i div 4)*4)];
		end;
	end;

	for i := 0 to 15 do
	begin
		aresult[i] := temp[i];
	end;
end;

procedure translateMatrix(var aresult: GLmatrix; x: GLfloat; y: GLfloat; z: GLfloat);
var
  matrix: GLmatrix;
  tempmatrix: GLmatrix;
begin
	loadIdentity(matrix);

	matrix[12] := x;
	matrix[13] := y;
	matrix[14] := z;

  tempmatrix := aresult;
	multMatrix(aresult, tempmatrix, matrix);
end;

procedure rotateRzRyRxMatrixf(var aresult: GLmatrix; anglex: GLfloat; angley: GLfloat; anglez: GLfloat);
var
  matrix : GLmatrix;
  tempmatrix : GLmatrix;
  rx : GLfloat;
	ry : GLfloat;
	rz : GLfloat;
	sx : GLfloat;
	cx : GLfloat;
	sy : GLfloat;
	cy : GLfloat;
	sz : GLfloat;
	cz : GLfloat;
begin
	rx := 2.0*PI*anglex/360.0;
	ry := 2.0*PI*angley/360.0;
	rz := 2.0*PI*anglez/360.0;
	sx := sin(rx);
	cx := cos(rx);
	sy := sin(ry);
	cy := cos(ry);
	sz := sin(rz);
	cz := cos(rz);

	loadIdentity(matrix);

	matrix[0] := cy*cz;
	matrix[1] := cy*sz;
	matrix[2] := -sy;

	matrix[4] := cz*sx*sy-cx*sz;
	matrix[5] := cx*cz+sx*sy*sz;
	matrix[6] := cy*sx;

	matrix[8] := cx*cz*sy+sx*sz;
	matrix[9] := -cz*sx+cx*sy*sz;
	matrix[10] := cx*cy;

  tempmatrix := aresult;
	multMatrix(aresult, tempmatrix, matrix);
end;

procedure frustum(var aresult: glmatrix; left: GLfloat; right: GLfloat; bottom: GLfloat; top: GLfloat; nearVal: GLfloat; farVal: GLfloat);
begin
	aresult[0] := 2.0*nearVal/(right-left);
	aresult[1] := 0.0;
	aresult[2] := 0.0;
	aresult[3] := 0.0;
	aresult[4] := 0.0;
	aresult[5] := 2.0*nearVal/(top-bottom);
	aresult[6] := 0.0;
	aresult[7] := 0.0;
	aresult[8] := (right+left)/(right-left);
	aresult[9] := (top+bottom)/(top-bottom);
	aresult[10] := -(farVal+nearVal)/(farVal-nearVal);
	aresult[11] := -1.0;
	aresult[12] := 0.0;
	aresult[13] := 0.0;
	aresult[14] := -(2.0*farVal*nearVal)/(farVal-nearVal);
	aresult[15] := 0.0;
end;

procedure perspective(var aresult: glmatrix; fovy: GLfloat; aspect: GLfloat; zNear: GLfloat; zFar: GLfloat);
var
  xmin, xmax, ymin, ymax: GLfloat;
begin
    ymax := zNear * tan(fovy * PI / 360.0);
    ymin := -ymax;
    xmin := ymin * aspect;
    xmax := ymax * aspect;

    frustum(aresult, xmin, xmax, ymin, ymax, zNear, zFar);
end;

procedure lookAt(var aresult: GLmatrix; eyeX: GLfloat; eyeY: GLfloat; eyeZ: GLfloat; centerX: GLfloat; centerY: GLfloat; centerZ: GLfloat; upX: GLfloat; upY: GLfloat; upZ: GLfloat );
var
  matrix: GLmatrix;
  tempmatrix : GLmatrix;
  fforward, side, up: GLvect;
begin
  fforward[0] := (centerX - eyeX);
  fforward[1] := (centerY - eyeY);
  fforward[2] := (centerZ - eyeZ);

  normalize(fforward);

  up[0] := upX;
  up[1] := upY;
  up[2] := upZ;

  crossProduct(side, fforward, up);
  normalize(side);

	crossProduct(up, side, fforward);

	matrix[0] := side[0];
	matrix[1] := up[0];
	matrix[2] := -fforward[0];
	matrix[3] := 0.0;
	matrix[4] := side[1];
	matrix[5] := up[1];
	matrix[6] := -fforward[1];
	matrix[7] := 0.0;
	matrix[8] := side[2];
	matrix[9] := up[2];
	matrix[10] := -fforward[2];
	matrix[11] := 0.0;
	matrix[12] := 0.0;
	matrix[13] := 0.0;
	matrix[14] := 0.0;
	matrix[15] := 1.0;

	loadIdentity(tempmatrix);
	multMatrix(aresult, tempmatrix, matrix);

	translateMatrix(aresult, -eyeX, -eyeY, -eyeZ);
end;

{------------------------------------------------------------------}
{  Function to draw the actual scene                               }
{------------------------------------------------------------------}
procedure glDraw();
begin
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);

  glBindVertexArray(vao);		// select first VAO

	// http://www.opengl.org/sdk/docs/man/xhtml/glDrawElements.xml
	glDrawElements(GL_TRIANGLES, 6*2*3, GL_UNSIGNED_INT, nil);

  glBindVertexArray(0);

end;


{------------------------------------------------------------------}
{  Initialise OpenGL                                               }
{------------------------------------------------------------------}
procedure glInit();
var
  g_vertices: GLuint; //The VBO for the vertices.
  g_normals: GLuint; //The VBO for the normals.

	g_modelViewLocation: GLint;
	g_vertexLocation: GLint;
	g_normalLocation: GLint;
  //Move the cube along the minus z axis for making it visible. Also rotated.
  g_modelView: GLmatrix;
  // Matrix for the model
	model: GLmatrix;
  tempmatrix: GLmatrix;

begin
  //load simple glsl shader
  glslsimpleprog := TGLSLProgram.Create();
  glslsimplevert := TGLSLShader.Create('shaders\Vertex.vs');
  log.Writeln(glslsimplevert.Log);
  glslsimplefrag := TGLSLShader.Create('shaders\Fragment.fs', GL_FRAGMENT_SHADER_ARB);
  log.Writeln(glslsimplefrag.Log);

  glslsimpleprog.Attach(glslsimplevert);
  log.Writeln(glslsimplevert.Log);
  glslsimpleprog.Attach(glslsimplefrag);
  log.Writeln(glslsimplefrag.Log);
  glslsimpleprog.Link;
  log.Writeln(glslsimpleprog.Log);
  glslsimpleprog.Enable;
  log.Writeln(glslsimpleprog.Log);

  // http://www.opengl.org/sdk/docs/man/xhtml/glGetUniformLocation.xml
	g_projectionLocation := glslsimpleprog.GetUniformLocation('projectionMatrix');

	// http://www.opengl.org/sdk/docs/man/xhtml/glGetUniformLocation.xml
	g_modelViewLocation := glslsimpleprog.GetUniformLocation('modelViewMatrix');

	// http://www.opengl.org/sdk/docs/man/xhtml/glGetAttribLocation.xml
	g_vertexLocation := glslsimpleprog.GetAttribLocation('vertex');

	// http://www.opengl.org/sdk/docs/man/xhtml/glGetAttribLocation.xml
	g_normalLocation := glslsimpleprog.GetAttribLocation('normal');

  log.Writeln( 'projectionLocation: '+inttostr(g_projectionLocation) );
  log.Writeln( 'modelviewLocation: '+inttostr(g_modelviewLocation) );
  log.Writeln( 'vertexLocation: '+inttostr(g_vertexLocation) );
  log.Writeln( 'normalLocation: '+inttostr(g_normalLocation) );

  // ToDo:
	glslsimpleprog.BindFragDataLocation('fragColor',0);

  // One VAO allocation's
  glGenVertexArrays(1,@vao);

  log.Writeln( 'vao id: '+inttostr(vao) );

  // First VAO setup
  glBindVertexArray(vao);

  glGenBuffers(1, @g_vertices);
  log.Writeln( 'g_vertices: '+inttostr(g_vertices) );
  glBindBuffer(GL_ARRAY_BUFFER, g_vertices);
  glBufferData(GL_ARRAY_BUFFER,24*4*sizeof(GLfloat), @points, GL_STATIC_DRAW);

  glGenBuffers(1, @g_normals);
  log.Writeln( 'g_normals: '+inttostr(g_normals) );
  glBindBuffer(GL_ARRAY_BUFFER, g_normals);
  glBufferData(GL_ARRAY_BUFFER,24*3*sizeof(GLfloat), @normals, GL_STATIC_DRAW);

  glGenBuffers(1, @g_indices);
  log.Writeln( 'g_indices: '+inttostr(g_indices) );
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, g_indices);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, 6*2*3*sizeof(GLuint), @indices, GL_STATIC_DRAW);

  // Calculate the model matrix ...
  loadIdentity(model);
	rotateRzRyRxMatrixf(model, 30.0, 30.0, 0.0);
	// ... and the view matrix ...
	lookAt(tempmatrix, 0.0, 0.0, 5.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
	// ... to get the final model view matrix
	multMatrix(g_modelView, tempmatrix, model);

  // http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml
	glUniformMatrix4fv(g_modelViewLocation, 1, false, @g_modelView);

  // http://www.opengl.org/sdk/docs/man/xhtml/glBindBuffer.xml
	glBindBuffer(GL_ARRAY_BUFFER, g_vertices);

	// http://www.opengl.org/sdk/docs/man/xhtml/glVertexAttribPointer.xml
	glVertexAttribPointer(g_vertexLocation, 4, GL_FLOAT, false, 0, 0);

	// http://www.opengl.org/sdk/docs/man/xhtml/glEnableVertexAttribArray.xml
	glEnableVertexAttribArray(g_vertexLocation);

	// http://www.opengl.org/sdk/docs/man/xhtml/glBindBuffer.xml
	glBindBuffer(GL_ARRAY_BUFFER, g_normals);

	// http://www.opengl.org/sdk/docs/man/xhtml/glVertexAttribPointer.xml
	glVertexAttribPointer(g_normalLocation, 3, GL_FLOAT, false, 0, 0);

	// http://www.opengl.org/sdk/docs/man/xhtml/glEnableVertexAttribArray.xml
	glEnableVertexAttribArray(g_normalLocation);

  glBindVertexArray(0); //Done creating the vao

	// http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/clearcolor.html
	glClearColor(0.0, 0.0, 0.0, 0.0);

	// http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/cleardepth.html
	glClearDepth(1.0);

	//http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/enable.html
	glEnable(GL_DEPTH_TEST);

	//http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/enable.html
	glEnable(GL_CULL_FACE);

  //glColor4f(1,1,1,1);
end;


{------------------------------------------------------------------}
{  Handle window resize                                            }
{------------------------------------------------------------------}
procedure glResizeWnd(Width, Height : Integer);
var
  //Projection matrix will be calculated later.
  g_projection: GLMatrix;
begin
  if (Height=0) then
     Height:=1;
  glViewport(0, 0, Width, Height);

  // Calculate the projection matrix and set it
	perspective(g_projection, 40.0, Width/Height, 1.0, 100.0);

	// http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml
	glUniformMatrix4fv(g_projectionLocation, 1, false, @g_projection);
end;

{------------------------------------------------------------------}
{  Processes all the mouse move                                  }
{------------------------------------------------------------------}
procedure ProcessMouseMove();
begin
  if MouseMove then
  begin
    // Add Code here to execute on mouse move
    MouseMove := False;
  end;
end;

{------------------------------------------------------------------}
{  Processes all the mouse clicks                                  }
{------------------------------------------------------------------}
procedure ProcessMouseClick;
begin
  case MouseButton of
  1: // Left Mouse Button
    begin
      // Add code to execute on mouse click
      MouseButton := 0; // Cancel our mouse click (To use this procedure as a mouse down event remove this line)
    end;
  2: // Right Mouse Button
    begin
      // Add code to execute on mouse click
      MouseButton := 0; // Cancel our mouse click (To use this procedure as a mouse down event remove this line)
    end;
  3: // Middle Mouse Button
    begin
      // Add code to execute on mouse click
      MouseButton := 0;  // Cancel our mouse click (To use this procedure as a mouse down event remove this line)
    end;
  end;
end;

{------------------------------------------------------------------}
{  Processes all the keystrokes                                    }
{------------------------------------------------------------------}
procedure ProcessKeys;
begin
  if (keys[VK_UP])    then xspeed := xspeed - 0.002;
  if (keys[VK_DOWN])  then xspeed := xspeed + 0.002;
  if (keys[VK_RIGHT]) then yspeed := yspeed + 0.002;
  if (keys[VK_LEFT])  then yspeed := yspeed - 0.002;
end;


{------------------------------------------------------------------}
{  Determines the applications response to the messages received  }
{------------------------------------------------------------------}
function WndProc(hWnd: HWND; Msg: UINT;  wParam: WPARAM;  lParam: LPARAM): LRESULT; stdcall;
begin
  case (Msg) of
    WM_CREATE:
      begin
        // Insert stuff you want executed when the program starts
      end;
    WM_CLOSE:
      begin
        PostQuitMessage(0);
        Result := 0
      end;
    WM_KEYDOWN:       // Set the pressed key (wparam) to equal true so we can check if its pressed
      begin
        keys[wParam] := True;
        Result := 0;
      end;
    WM_KEYUP:         // Set the released key (wparam) to equal false so we can check if its pressed
      begin
        keys[wParam] := False;
        Result := 0;
      end;
    WM_MOUSEMOVE:
      begin
        MouseMove := True;
        MoveXcoord := LOWORD(lParam);
        MoveYcoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_LBUTTONDOWN:
      begin
        ReleaseCapture();   // need them here, because if mouse moves off
        SetCapture(h_Wnd);  // window and returns, it needs to reset status
        MouseButton := 1;
        Xcoord := LOWORD(lParam);
        Ycoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_RBUTTONDOWN:
      begin
        ReleaseCapture();   // need them here, because if mouse moves off
        SetCapture(h_Wnd);  // window and returns, it needs to reset status
        MouseButton := 2;
        Xcoord := LOWORD(lParam);
        Ycoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_MBUTTONDOWN:
      begin
        ReleaseCapture();   // need them here, because if mouse moves off
        SetCapture(h_Wnd);  // window and returns, it needs to reset status
        MouseButton := 3;
        Xcoord := LOWORD(lParam);
        Ycoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_LBUTTONUP,WM_RBUTTONUP,WM_MBUTTONUP:
      begin
        ReleaseCapture();   // above
        MouseButton := 0;
        XCoord := 0;
        YCoord := 0;
        Result := 0;
      end;
    WM_SIZE:          // Resize the window with the new width and height
      begin
        if OpenGLInitialized then
         glResizeWnd(LOWORD(lParam),HIWORD(lParam));
        Result := 0;
      end;
    WM_TIMER :                     // Add code here for all timers to be used.
      begin
        if wParam = FPS_TIMER then
        begin
          FPSCount :=Round(FPSCount * 1000/FPS_INTERVAL);   // calculate to get per Second incase intercal is less or greater than 1 second
          SetWindowText(h_Wnd, PChar(WND_TITLE + '   [' + intToStr(FPSCount) + ' FPS]'));
          FPSCount := 0;
          Result := 0;
        end;
      end;
    else
      begin

      Result := DefWindowProc(hWnd, Msg, wParam, lParam);    // Default result if nothing happens
      end;
  end;
end;


{---------------------------------------------------------------------}
{  Properly destroys the window created at startup (no memory leaks)  }
{---------------------------------------------------------------------}
procedure glKillWnd(Fullscreen : Boolean);
begin
  if Fullscreen then             // Change back to non fullscreen
  begin
    ChangeDisplaySettings(devmode(nil^), 0);
    ShowCursor(True);
  end;

  // Makes current rendering context not current, and releases the device
  // context that is used by the rendering context.
  if (not wglMakeCurrent(h_DC, 0)) then
    MessageBox(0, 'Release of DC and RC failed!', 'Error', MB_OK or MB_ICONERROR);

  // Attempts to delete the rendering context normal or OpenGL 3.x
  if OpenGL3Initialized then
  begin
    if (not wglDeleteContext(h_RC3)) then
    begin
      MessageBox(0, 'Release of rendering context failed!', 'Error', MB_OK or MB_ICONERROR);
      h_RC3 := 0;
    end;
  end else
  begin
    if (not wglDeleteContext(h_RC)) then
    begin
      MessageBox(0, 'Release of rendering context failed!', 'Error', MB_OK or MB_ICONERROR);
      h_RC := 0;
    end;
  end;

  // Attemps to release the device context
  if ((h_DC > 0) and (ReleaseDC(h_Wnd, h_DC) = 0)) then
  begin
    MessageBox(0, 'Release of device context failed!', 'Error', MB_OK or MB_ICONERROR);
    h_DC := 0;
  end;

  // Attempts to destroy the window
  if ((h_Wnd <> 0) and (not DestroyWindow(h_Wnd))) then
  begin
    MessageBox(0, 'Unable to destroy window!', 'Error', MB_OK or MB_ICONERROR);
    h_Wnd := 0;
  end;

  // Attempts to unregister the window class
  if (not UnRegisterClass('OpenGL', hInstance)) then
  begin
    MessageBox(0, 'Unable to unregister window class!', 'Error', MB_OK or MB_ICONERROR);
    hInstance := 0;
  end;
end;


{--------------------------------------------------------------------}
{  Creates the window and attaches a OpenGL rendering context to it  }
{--------------------------------------------------------------------}
function glCreateWnd(Width, Height : Integer; Fullscreen : Boolean; PixelDepth : Integer) : Boolean;
var
  wndClass : TWndClass;         // Window class
  dwStyle : DWORD;              // Window styles
  dwExStyle : DWORD;            // Extended window styles
  dmScreenSettings : DEVMODE;   // Screen settings (fullscreen, etc...)
  PixelFormat : GLuint;         // Settings for the OpenGL rendering
  h_Instance : HINST;           // Current instance
  pfd : TPIXELFORMATDESCRIPTOR;  // Settings for the OpenGL window
  attribs: array[0..6] of integer; //OpenGL 3.x Attribs
begin
  InitOpenGL;                               // New call to initialize and bind the OpenGL dll

  h_Instance := GetModuleHandle(nil);       //Grab An Instance For Our Window
  ZeroMemory(@wndClass, SizeOf(wndClass));  // Clear the window class structure

  with wndClass do                    // Set up the window class
  begin
    style         := CS_HREDRAW or    // Redraws entire window if length changes
                     CS_VREDRAW or    // Redraws entire window if height changes
                     CS_OWNDC;        // Unique device context for the window
    lpfnWndProc   := @WndProc;        // Set the window procedure to our func WndProc
    hInstance     := h_Instance;
    hCursor       := LoadCursor(0, IDC_ARROW);
    lpszClassName := 'OpenGL';
  end;

  if (RegisterClass(wndClass) = 0) then  // Attemp to register the window class
  begin
    MessageBox(0, 'Failed to register the window class!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit
  end;

  // Change to fullscreen if so desired
  if Fullscreen then
  begin
    ZeroMemory(@dmScreenSettings, SizeOf(dmScreenSettings));
    with dmScreenSettings do begin              // Set parameters for the screen setting
      dmSize       := SizeOf(dmScreenSettings);
      dmPelsWidth  := Width;                    // Window width
      dmPelsHeight := Height;                   // Window height
      dmBitsPerPel := PixelDepth;               // Window color depth
      dmFields     := DM_PELSWIDTH or DM_PELSHEIGHT or DM_BITSPERPEL;
    end;

    // Try to change screen mode to fullscreen
    if (ChangeDisplaySettings(dmScreenSettings, CDS_FULLSCREEN) = DISP_CHANGE_FAILED) then
    begin
      MessageBox(0, 'Unable to switch to fullscreen!', 'Error', MB_OK or MB_ICONERROR);
      Fullscreen := False;
    end;
  end;

  // If we are still in fullscreen then
  if (Fullscreen) then
  begin
    dwStyle := WS_POPUP or                // Creates a popup window
               WS_CLIPCHILDREN            // Doesn't draw within child windows
               or WS_CLIPSIBLINGS;        // Doesn't draw within sibling windows
    dwExStyle := WS_EX_APPWINDOW;         // Top level window
    ShowCursor(False);                    // Turn of the cursor (gets in the way)
  end
  else
  begin
    dwStyle := WS_OVERLAPPEDWINDOW or     // Creates an overlapping window
               WS_CLIPCHILDREN or         // Doesn't draw within child windows
               WS_CLIPSIBLINGS;           // Doesn't draw within sibling windows
    dwExStyle := WS_EX_APPWINDOW or       // Top level window
                 WS_EX_WINDOWEDGE;        // Border with a raised edge
  end;

  // Attempt to create the actual window
  h_Wnd := CreateWindowEx(dwExStyle,      // Extended window styles
                          'OpenGL',       // Class name
                          WND_TITLE,      // Window title (caption)
                          dwStyle,        // Window styles
                          0, 0,           // Window position
                          Width, Height,  // Size of window
                          0,              // No parent window
                          0,              // No menu
                          h_Instance,     // Instance
                          nil);           // Pass nothing to WM_CREATE

  if h_Wnd = 0 then
  begin
    glKillWnd(Fullscreen);                // Undo all the settings we've changed
    MessageBox(0, 'Unable to create window!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Try to get a device context
  h_DC := GetDC(h_Wnd);
  if (h_DC = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to get a device context!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Settings for the OpenGL window
  with pfd do
  begin
    nSize           := SizeOf(TPIXELFORMATDESCRIPTOR); // Size Of This Pixel Format Descriptor
    nVersion        := 1;                    // The version of this data structure
    dwFlags         := PFD_DRAW_TO_WINDOW    // Buffer supports drawing to window
                       or PFD_SUPPORT_OPENGL // Buffer supports OpenGL drawing
                       or PFD_DOUBLEBUFFER;  // Supports double buffering
    iPixelType      := PFD_TYPE_RGBA;        // RGBA color format
    cColorBits      := PixelDepth;           // OpenGL color depth
    cRedBits        := 0;                    // Number of red bitplanes
    cRedShift       := 0;                    // Shift count for red bitplanes
    cGreenBits      := 0;                    // Number of green bitplanes
    cGreenShift     := 0;                    // Shift count for green bitplanes
    cBlueBits       := 0;                    // Number of blue bitplanes
    cBlueShift      := 0;                    // Shift count for blue bitplanes
    cAlphaBits      := 0;                    // Not supported
    cAlphaShift     := 0;                    // Not supported
    cAccumBits      := 0;                    // No accumulation buffer
    cAccumRedBits   := 0;                    // Number of red bits in a-buffer
    cAccumGreenBits := 0;                    // Number of green bits in a-buffer
    cAccumBlueBits  := 0;                    // Number of blue bits in a-buffer
    cAccumAlphaBits := 0;                    // Number of alpha bits in a-buffer
    cDepthBits      := PixelDepth;           // Specifies the depth of the depth buffer
    cStencilBits    := 0;                    // Specifies the depth of the stencil buffer
    cAuxBuffers     := 0;                    // Not supported
    iLayerType      := PFD_MAIN_PLANE;       // Ignored
    bReserved       := 0;                    // Number of overlay and underlay planes
    dwLayerMask     := 0;                    // Ignored
    dwVisibleMask   := 0;                    // Transparent color of underlay plane
    dwDamageMask    := 0;                     // Ignored
  end;

  // Attempts to find the pixel format supported by a device context that is the best match to a given pixel format specification.
  PixelFormat := ChoosePixelFormat(h_DC, @pfd);
  if (PixelFormat = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to find a suitable pixel format', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Sets the specified device context's pixel format to the format specified by the PixelFormat.
  if (not SetPixelFormat(h_DC, PixelFormat, @pfd)) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to set the pixel format', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Create a OpenGL rendering context
  h_RC := wglCreateContext(h_DC);
  if (h_RC = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to create an OpenGL rendering context', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Makes the specified OpenGL rendering context the calling thread's current rendering context
  if (not wglMakeCurrent(h_DC, h_RC)) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to activate OpenGL rendering context', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Read And Assign Extentions
  ReadExtensions;
  ReadImplementationProperties;

  if dglCheckExtension('WGL_ARB_create_context') then
  begin

    // Set OpenGL 3.x attribs
    attribs[0]:=WGL_CONTEXT_MAJOR_VERSION_ARB;
    attribs[1]:=3;
    attribs[2]:=WGL_CONTEXT_MINOR_VERSION_ARB;
    attribs[3]:=0;
    attribs[4]:=WGL_CONTEXT_FLAGS_ARB;
    attribs[5]:=WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB;
    attribs[6]:=0;

    // Create OpenGL 3.x context using the attribs
    h_RC3 := wglCreateContextAttribsARB(h_DC, 0, @attribs);
    if h_RC3 <> 0 then
    begin
      wglMakeCurrent(0,0); //remove current context
      wglDeleteContext(h_RC); //remove opengl 2.x context
      wglMakeCurrent(h_DC, h_RC3); //make the opengl 3.x context current
      OpenGL3Initialized := True;
    end;
  end else
  begin
    glKillWnd(Fullscreen); //If you do not use OpenGL 3.x specific function you can also continue here ...
    MessageBox(0, 'It is not possible to create an OpenGL 3.x context', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  OpenGLInitialized := True;
  // Initializes the timer used to calculate the FPS
  SetTimer(h_Wnd, FPS_TIMER, FPS_INTERVAL, nil);

  // Settings to ensure that the window is the topmost window
  ShowWindow(h_Wnd, SW_SHOW);
  SetForegroundWindow(h_Wnd);
  SetFocus(h_Wnd);

  // Ensure the OpenGL window is resized properly
  glResizeWnd(Width, Height);
  glInit();

  Result := True;
end;


{--------------------------------------------------------------------}
{  Main message loop for the application                             }
{--------------------------------------------------------------------}
function WinMain(hInstance : HINST; hPrevInstance : HINST;
                 lpCmdLine : PChar; nCmdShow : Integer) : Integer; stdcall;
var
  msg : TMsg;
  finished : Boolean;
  DemoStart, LastTime : DWord;
begin
  finished := False;

  // Perform application initialization:
  if not glCreateWnd(winwidth, winheight, FALSE, 32) then
  begin
    Result := 0;
    Exit;
  end;

  DemoStart := GetTickCount();            // Get Time when demo started

  // Main message loop:
  while not finished do
  begin
    if (PeekMessage(msg, 0, 0, 0, PM_REMOVE)) then // Check if there is a message for this window
    begin
      if (msg.message = WM_QUIT) then     // If WM_QUIT message received then we are done
        finished := True
      else
      begin                               // Else translate and dispatch the message to this window
  	TranslateMessage(msg);
        DispatchMessage(msg);
      end;
    end
    else
    begin
      Inc(FPSCount);                      // Increment FPS Counter

      LastTime :=ElapsedTime;
      ElapsedTime :=GetTickCount() - DemoStart;     // Calculate Elapsed Time
      ElapsedTime :=(LastTime + ElapsedTime) DIV 2; // Average it out for smoother movement

      glDraw();                           // Draw the scene
      SwapBuffers(h_DC);                  // Display the scene

      if (keys[VK_ESCAPE]) then           // If user pressed ESC then set finised TRUE
        finished := True
      else
      begin
        ProcessKeys;                      // Check for any other key Pressed
        ProcessMouseMove;                 // Check for mouse move
        ProcessMouseClick;                // Check for Mouse Click
      end;
    end;
  end;
  glKillWnd(FALSE);
  Result := msg.wParam;
end;


begin
  WinMain( hInstance, hPrevInst, CmdLine, CmdShow );
end.

